[PATCH] KEYS: fix refcount_inc() on zero

Mark Rutland mark.rutland at arm.com
Fri May 26 17:37:34 UTC 2017


If a key's refcount is dropped to zero between key_lookup() peeking at
the refcount and subsequently attempting to increment it, refcount_inc()
will see a zero refcount. Here, refcount_inc() will WARN_ONCE(), and
will *not* increment the refcount, which will remain zero.

Once key_lookup() drops key_serial_lock, it is possible for the key to
be freed behind our back.

This patch uses refcount_inc_not_zero() to perform the peek and
increment atomically. A helper with lockdep annotation is added to
document why this is safe.

Fixes: fff292914d3a2f1e ("security, keys: convert key.usage from atomic_t to refcount_t")
Signed-off-by: Mark Rutland <mark.rutland at arm.com>
Cc: David Howells <dhowells at redhat.com>
Cc: David Windsor <dwindsor at gmail.com>
Cc: Elena Reshetova <elena.reshetova at intel.com>
Cc: Hans Liljestrand <ishkamiel at gmail.com>
Cc: James Morris <james.l.morris at oracle.com>
Cc: Kees Cook <keescook at chromium.org>
Cc: Peter Zijlstra <peterz at infradead.org>
---
 security/keys/key.c | 18 +++++++++++-------
 1 file changed, 11 insertions(+), 7 deletions(-)

diff --git a/security/keys/key.c b/security/keys/key.c
index 455c04d..150f51d 100644
--- a/security/keys/key.c
+++ b/security/keys/key.c
@@ -632,6 +632,12 @@ void key_put(struct key *key)
 }
 EXPORT_SYMBOL(key_put);
 
+static bool key_get_not_free(struct key *key)
+{
+	lockdep_assert_held(&key_serial_lock);
+	return refcount_inc_not_zero(&key->usage);
+}
+
 /*
  * Find a key by its serial number.
  */
@@ -660,14 +666,12 @@ struct key *key_lookup(key_serial_t id)
 	goto error;
 
 found:
-	/* pretend it doesn't exist if it is awaiting deletion */
-	if (refcount_read(&key->usage) == 0)
-		goto not_found;
-
-	/* this races with key_put(), but that doesn't matter since key_put()
-	 * doesn't actually change the key
+	/*
+	 * Pretend it doesn't exist if it is awaiting deletion. This races with
+	 * key_put(), but we can peek at the key until we drop key_serial_lock.
 	 */
-	__key_get(key);
+	if (!key_get_not_free(key))
+		goto not_found;
 
 error:
 	spin_unlock(&key_serial_lock);
-- 
1.9.1

--
To unsubscribe from this list: send the line "unsubscribe linux-security-module" in
the body of a message to majordomo at vger.kernel.org
More majordomo info at  http://vger.kernel.org/majordomo-info.html



More information about the Linux-security-module-archive mailing list